from pydantic import BaseModel, Field

from .checks import CheckModel


class HealthResponse(BaseModel):
    name: str = Field(..., description='The name of the Service')
    version: str = Field(..., description='The Version of the Service')
    all_ok: bool = Field(..., description='If all checks are OK')
    checks: list[CheckModel] = Field(..., description='List of the Checks executed')



import asyncio
import inspect
import logging
import os
from logging.config import dictConfig
from pathlib import Path

from edgegap_logging import Color, Format
from edgegap_scheduling import Scheduler
from edgegap_settings import ProjectBaseSettings
from elasticapm.contrib import starlette
from fastapi import FastAPI, staticfiles, templating

from ._configuration import ServiceConfiguration
from ._documentation import DocumentationRoute
from ._scheduling import SchedulingAPI
from ._templating import TemplateRenderer
from .health import HealthCheck

current_dir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))


class ServiceFactory:
    def __init__(self, configuration: ServiceConfiguration) -> None:
        self.__configuration = configuration
        self.__static_folder = os.path.join(self.__configuration.root_dir, self.__configuration.static_dir)
        self.__logger: logging.Logger | None = None
        self.__app: FastAPI = None

    def create_app(self) -> FastAPI:
        self.__app = FastAPI(
            title=self.__configuration.name,
            description=self.__configuration.description,
            configuration=self.__configuration,
            openapi_tags=self.__configuration.tags_definition,
        )

        Path(self.__static_folder).mkdir(parents=True, exist_ok=True)

        for router in self.__configuration.routers:
            self.__app.include_router(router)

        self.__app.mount(
            '/static',
            staticfiles.StaticFiles(
                directory=self.__static_folder,
                packages=[('edgegap_service', 'static')],
            ),
            name='static',
        )

        self.__logger = self.__init_logging()

        service_string = f'{self.__configuration.name} ({self.__configuration.version})'
        self.__logger.info(
            f'Service {Format.squared(service_string, Color.GREEN)} is initializing',
        )

        self.__init_root()
        self.__init_health()
        self.__init_template_renderer()
        self.__init_apm()
        self.__init_startup_functions()
        self.__init_scheduling()
        self.__init_sigterm()

        return self.__app

    def __init_root(self):
        folder = os.path.join(current_dir, 'static', 'html')
        file_name = 'index.html'
        is_defined = isinstance(self.__configuration.html_index_path, str)

        if is_defined:
            full_path = os.path.join(self.__configuration.root_dir, self.__configuration.html_index_path)

            if os.path.isfile(full_path):
                folder = os.path.dirname(full_path)
                file_name = os.path.basename(full_path)

        templates = templating.Jinja2Templates(directory=folder)
        DocumentationRoute.init_documentation(self.__app, file_name, templates)
        self.__logger.info(
            f'Initialized root path and documentation with folder: {Format.squared(folder, Color.GREEN)}',
        )

    def __init_health(self):
        health_check = HealthCheck(
            title=self.__configuration.name,
            version=self.__configuration.version,
            checks=self.__configuration.checks,
            depend=self.__configuration.depend,
        )
        health_check.init_health_check(self.__app)

        self.__logger.info(f"Initialized Health route at {Format.squared('/health', Color.GREEN)}")

    def __init_template_renderer(self):
        assert isinstance(
            TemplateRenderer.get_templates(configuration=self.__configuration),
            templating.Jinja2Templates,
        )
        self.__logger.info('Initialized Template Renderer')

    def __init_logging(self) -> logging.Logger:
        self.__init_logstash()

        dictConfig(self.__configuration.log_config.model_dump())
        logger = logging.getLogger(self.__configuration.name)
        logger.info(f'Logging for service {Format.squared(self.__configuration.name, Color.GREEN)} is configured')

        if (
            isinstance(self.__configuration.settings, ProjectBaseSettings)
            and self.__configuration.settings.logstash.enabled
        ):
            logger.info('Logstash is enabled')

        return logger

    def __init_apm(self):
        if isinstance(self.__configuration.settings, ProjectBaseSettings):
            apm = self.__configuration.settings.apm

            if apm.enabled:
                self.__logger.info('Apm is enabled')
                client = starlette.make_apm_client({
                    'SERVICE_NAME': self.__configuration.name,
                    'SECRET_TOKEN': apm.token,
                    'SERVER_URL': f'{apm.scheme}://{apm.server}:{apm.port}',
                })
                data = {
                    'middleware_class': starlette.ElasticAPM,
                    'client': client,
                }
                self.__app.add_middleware(**data)

    def __init_logstash(self):
        if isinstance(self.__configuration.settings, ProjectBaseSettings):
            logstash = self.__configuration.settings.logstash

            if logstash.enabled:
                formatter = {
                    'logstash': {
                        'class': 'logstash_async.formatter.LogstashFormatter',
                        'fmt': self.__configuration.env_config.log_format,
                    },
                }
                handlers = {
                    'logstash': {
                        'class': 'logstash_async.handler.AsynchronousLogstashHandler',
                        'formatter': 'logstash',
                        'transport': 'logstash_async.transport.TcpTransport',
                        'args': "('%(host)s', %(port)s, '%(database_path)s', '%(transport)s', "
                        "%(ssl_enable)s, %(ssl_verify)s, '%(keyfile)s', '%(certfile)s', "
                        "'%(ca_certs)s', %(enable)s)",
                        'host': logstash.server,
                        'port': logstash.port,
                        'enable': logstash.enabled,
                        'ssl_enable': False,
                        'ssl_verify': False,
                        'database_path': None,
                    },
                }
                self.__configuration.log_config.root['handlers'] = ['default', 'logstash']
                self.__configuration.log_config.loggers['root']['handlers'] = ['default', 'logstash']
                self.__configuration.log_config.formatters.update(formatter)
                self.__configuration.log_config.handlers.update(handlers)

    def __init_scheduling(self):
        if isinstance(self.__configuration.scheduler, Scheduler):

            @self.__app.on_event('startup')
            def start_scheduling():
                SchedulingAPI().init_scheduling_api(self.__app)
                asyncio.create_task(self.__configuration.scheduler.start_all())

    def __init_sigterm(self):
        @self.__app.on_event('shutdown')
        async def handle():
            try:
                self.__logger.info(f'Shutting down {Format.squared(self.__configuration.name, Color.GREEN)}')

                if isinstance(self.__configuration.scheduler, Scheduler):
                    asyncio.ensure_future(self.__configuration.scheduler.stop_all())
            except Exception as e:
                self.__logger.exception(f'Exception while shutting down: {e}')

            for func in self.__configuration.shutdown_functions:
                try:
                    if inspect.iscoroutinefunction(func):
                        await func()
                    else:
                        func()
                except Exception as e:
                    self.__logger.exception(f'Could not shutdown correctly one of the shutdown functions: {e}')

    def __init_startup_functions(self):
        @self.__app.on_event('startup')
        async def init_startup_functions():
            for func in self.__configuration.startup_functions:
                try:
                    if inspect.iscoroutinefunction(func):
                        await func()
                    else:
                        func()
                except Exception as e:
                    self.__logger.exception(f'Could not initialize correctly one of the startup functions: {e}')



from ._consul import ConsulCheck
from ._database import DatabaseCheck
from ._interface import CheckInterface
from ._model import CheckInstanceModel
from ._model import CheckModel

__all__ = [
    'CheckInterface',
    'CheckModel',
    'CheckInstanceModel',
    'ConsulCheck',
    'DatabaseCheck',
]



from edgegap_consul import ConsulReaderFactory

from ._interface import CheckInterface
from ._model import CheckInstanceModel


class ConsulCheck(CheckInterface):
    __name__ = 'Consul'

    def __init__(self, name: str = None):
        super().__init__(name)
        self.__reader = ConsulReaderFactory().from_env()

    def _check(self, **kwargs) -> CheckInstanceModel:
        try:
            ok = self.__reader.check()
            message = 'ok'
        except Exception as e:
            ok = False
            message = str(e)

        return CheckInstanceModel(ok=ok, message=message)



import os

from fastapi import templating

from ._configuration import ServiceConfiguration


class TemplateRenderer:
    __templates: templating.Jinja2Templates | None = None

    @classmethod
    def get_templates(cls, configuration: ServiceConfiguration = None) -> templating.Jinja2Templates:
        if isinstance(configuration, ServiceConfiguration) and cls.__templates is None:
            static_folder = os.path.join(configuration.root_dir, configuration.static_dir)
            cls.__templates = templating.Jinja2Templates(directory=static_folder)

        if not isinstance(cls.__templates, templating.Jinja2Templates):
            raise Exception(
                'Templating is not a Jinja2Templates instance, please pass a valid ServiceConfiguration first',
            )

        return cls.__templates

    @classmethod
    def depends_templates(cls) -> templating.Jinja2Templates:
        return cls.get_templates()

    @classmethod
    def clear_templates(cls) -> None:
        cls.__templates = None



import os
from typing import Callable, Optional

from edgegap_scheduling import Scheduler
from edgegap_settings import ProjectBaseSettings, SettingsFactory
from fastapi import APIRouter
from pydantic import BaseModel, ConfigDict, Field
from pydantic_settings import BaseSettings

from ._environment import EnvironmentConfiguration
from .health.checks import CheckInterface
from .logging import LoggingConfiguration


def get_default_root_dir() -> str:
    return os.getcwd()


class ServiceConfiguration(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    name: str = Field(..., description='The name of the Service')
    version: str = Field(..., description='The version of the Service')
    description: str = Field(..., description='The description of the Service')
    routers: list[APIRouter] = Field([], description='The FastAPI routers')
    root_dir: str = Field(default_factory=get_default_root_dir, description='The Root Directory of the Service')
    static_dir: str = Field(default='static', description='The Static Directory of the Service')
    html_index_path: str | None = Field(default=None, description='The full path to the html index file')
    checks: list[type(CheckInterface)] = Field([], description='The list of Checks to run on the Service')
    depend: Optional[Callable] = Field(default=None, description='The Base Depend Attribute for DB Session')
    port: int = Field(default=8000, description='The Port of the Service')
    host: str = Field(default='0.0.0.0', description='The Host of the Service')
    workers: int = Field(default=1, description='Numbers of Workers, (if you use the scheduler, keep to 1)')
    timeout: int = Field(default=300, description='Default Timeout for Connection')
    worker_class: str = Field(default='uvicorn.workers.UvicornWorker', description='The worker class for ASGI')
    scheduler: Scheduler | None = Field(default=None, description='A Scheduler to start at boot')
    startup_functions: list[Callable] = Field(default=[], description='All function will be called on startup')
    shutdown_functions: list[Callable] = Field(default=[], description='All function will be called on shutdown')
    env_config: EnvironmentConfiguration = Field(
        default=SettingsFactory.from_settings(EnvironmentConfiguration),
        description='The Configuration coming from the Environment Variables',
    )
    log_config: LoggingConfiguration = Field(
        default=LoggingConfiguration(),
        description='The Configuration for the Logging, will be passed as DictConfig to Python logging system',
    )
    settings: ProjectBaseSettings | BaseSettings | None = Field(
        default=None,
        description='Project Specific Settings to hold in the Service',
    )
    tags_definition: list[dict] = Field(
        default=[],
        description='See https://fastapi.tiangolo.com/tutorial/metadata/#use-your-tags for details',
    )



import http
import logging
from copy import copy

from edgegap_logging import Color, DefaultFormatter, Format


class AccessFormatter(DefaultFormatter):
    status_code_colours = {
        1: Color.WHITE,
        2: Color.GREEN,
        3: Color.YELLOW,
        4: Color.LIGHTRED_EX,
        5: Color.RED,
    }

    default_fmt = '%(asctime)s | %(levelname)s | %(name)s | %(client_addr)s - [%(request_line)s] %(status_code)s'

    def get_status_code(self, status_code: int) -> str:
        try:
            status_phrase = http.HTTPStatus(status_code).phrase
        except ValueError:
            status_phrase = ''

        status_and_phrase = f'{status_code} {status_phrase}'

        if self.use_colors:
            color = self.status_code_colours.get(status_code // 100)

            return Format.color(status_phrase, color)

        return status_and_phrase

    def formatMessage(self, record: logging.LogRecord) -> str:
        record_copy = copy(record)

        (
            client_addr,
            method,
            full_path,
            http_version,
            status_code,
        ) = record_copy.args  # type: ignore[misc]
        status_code = self.get_status_code(int(status_code))  # type: ignore[arg-type]
        request_line = f'{method} {full_path} HTTP/{http_version}'

        if self.use_colors:
            request_line = Format.color(request_line, bold=True)

        record_copy.__dict__.update(
            {
                'client_addr': client_addr,
                'request_line': request_line,
                'status_code': status_code,
            },
        )
        return super().formatMessage(record_copy)
